@@ -17,7 +17,7 @@
  * along with QCAD.
  */
 
-include("../Ellipse.js");
+include("scripts/Draw/Ellipse/Ellipse.js");
 include("scripts/ShapeAlgorithms.js");
 
 /**
@@ -243,6 +243,8 @@ EllipseInscribedQuad.prototype.getOperation = function(preview) {
         return undefined;
     }
 
+    shape = Ellipse.postProcess(shape);
+
     var doc = this.getDocument();
 
     var e = shapeToEntity(doc, shape);
