@@ -1,5 +1,5 @@
 /**
- * Copyright (c) 2011-2017 by Andrew Mustun. All rights reserved.
+ * Copyright (c) 2011-2018 by Andrew Mustun. All rights reserved.
  * 
  * This file is part of the QCAD project.
  *
@@ -17,7 +17,7 @@
  * along with QCAD.
  */
 
-include("../Widgets.js");
+include("scripts/Widgets/Widgets.js");
 include("scripts/WidgetFactory.js");
 
 /**
@@ -66,6 +66,7 @@ function RBlockListQt(parent, addListener, showHeader) {
         var adapter = new RBlockListenerAdapter();
         appWin.addBlockListener(adapter);
         adapter.blocksUpdated.connect(this, "updateBlocks");
+        adapter.currentBlockSet.connect(this, "updateCurrentBlock");
         adapter.blocksCleared.connect(this, "clearBlocks");
     }
 
@@ -73,10 +74,46 @@ function RBlockListQt(parent, addListener, showHeader) {
     this.itemColumnClicked.connect(this, "itemColumnClickedSlot");
     this.itemSelectionChanged.connect(this, "blockActivated");
     this.basePath = includeBasePath;
+
+    this.currentItem = undefined;
+
 }
 
 RBlockListQt.prototype = new RTreeWidget();
 
+RBlockListQt.getWidget = function() {
+    var appWin = EAction.getMainWindow();
+    return appWin.findChild("BlockList");
+};
+
+/**
+ * Add an additional action to be used in the context menu.
+ * Can be used by plugins to hook into the block list context menu.
+ */
+RBlockListQt.addContextMenuScriptFile = function(scriptFile, blockList) {
+    //var blockList = RBlockListQt.getWidget();
+    if (isNull(blockList)) {
+        return;
+    }
+
+    var scriptFiles = RBlockListQt.getContextMenuScriptFiles(blockList);
+    scriptFiles.push(scriptFile);
+    blockList.setProperty("ContextMenuScriptFiles", scriptFiles);
+};
+
+RBlockListQt.getContextMenuScriptFiles = function(blockList) {
+    //var blockList = RBlockListQt.getWidget();
+    if (isNull(blockList)) {
+        return [];
+    }
+
+    var scriptFiles = blockList.property("ContextMenuScriptFiles");
+    if (isNull(scriptFiles)) {
+        scriptFiles = [];
+    }
+    return scriptFiles;
+};
+
 RBlockListQt.prototype.contextMenuEvent = function(e) {
     var item = this.itemAt(e.pos());
     if (!isNull(item)) {
@@ -84,6 +121,7 @@ RBlockListQt.prototype.contextMenuEvent = function(e) {
     }
 
     var menu = new QMenu();
+
     RGuiAction.getByScriptFile("scripts/Block/ToggleBlockVisibility/ToggleBlockVisibility.js").addToMenu(menu);
     RGuiAction.getByScriptFile("scripts/Block/ShowAllBlocks/ShowAllBlocks.js").addToMenu(menu);
     RGuiAction.getByScriptFile("scripts/Block/HideAllBlocks/HideAllBlocks.js").addToMenu(menu);
@@ -95,11 +133,25 @@ RBlockListQt.prototype.contextMenuEvent = function(e) {
     RGuiAction.getByScriptFile("scripts/Block/SelectBlockReferences/SelectBlockReferences.js").addToMenu(menu);
     RGuiAction.getByScriptFile("scripts/Block/DeselectBlockReferences/DeselectBlockReferences.js").addToMenu(menu);
 
+    RBlockListQt.complementContextMenu(menu, RBlockListQt.getWidget());
+
     menu.exec(QCursor.pos());
 
     e.ignore();
 };
 
+RBlockListQt.complementContextMenu = function(menu, blockListWidget) {
+    // add addtional context menu actions provided by plugins:
+    var scriptFiles = RBlockListQt.getContextMenuScriptFiles(blockListWidget);
+    for (var i=0; i<scriptFiles.length; i++) {
+        var scriptFile = scriptFiles[i];
+        var a = RGuiAction.getByScriptFile(scriptFile);
+        if (!isNull(a)) {
+            a.addToMenu(menu);
+        }
+    }
+};
+
 RBlockListQt.prototype.filter = function(block) {
     var hideInternal = RSettings.getBoolValue("BlockList/HideInternalBlocks", false);
 
@@ -122,64 +174,38 @@ RBlockListQt.prototype.filter = function(block) {
     return true;
 };
 
-RBlockListQt.prototype.sortBlockNames = function(blockNames) {
-    var doc = this.di.getDocument();
+RBlockListQt.prototype.updateCurrentBlock = function(documentInterface) {
+    var doc = documentInterface.getDocument();
 
-    //var self = this;
-    blockNames.sort(function(a, b) {
-        var blockA = doc.queryBlock(a);
-        var blockB = doc.queryBlock(b);
-        if (isNull(blockA) || isNull(blockB)) {
-            return 0;
-        }
+    qDebug("updateCurrentBlock");
+    // find item which was representing the current block before:
+    if (!isNull(this.currentItem)) {
+        // clear pen icon from item:
+        //var blockName = this.currentItem.data(BlockList.colName, Qt.UserRole);
+        //var block = doc.queryBlock(blockName);
+        //this.updateItemIcons(this.currentItem, block);
+        this.currentItem.setIcon(BlockList.colEdit, BlockList.iconEdit[0]);
+    }
 
-        // model space always first:
-        if (blockA.getName()===RBlock.modelSpaceName) {
-            return -1;
-        }
-        if (blockB.getName()===RBlock.modelSpaceName) {
-            return 1;
-        }
+    // find item of current block:
+    this.currentItem = RBlockListQt.getItem(this, doc.getCurrentBlockName());
+    //for (var i=0; i<this.topLevelItemCount; i++) {
+    //    var item = this.topLevelItem(i);
+    //}
 
-        // layouts always before other blocks:
-        var blockAIsLayout = blockA.isLayout();
-        var blockBIsLayout = blockB.isLayout();
-        if (blockAIsLayout && !blockBIsLayout) {
-            return -1;
-        }
-        if (!blockAIsLayout && blockBIsLayout) {
-            return 1;
-        }
-
-        // sort among layouts by tab order:
-        if (blockAIsLayout && blockBIsLayout) {
-            var layoutAId = blockA.getLayoutId();
-            var layoutBId = blockB.getLayoutId();
-            var layoutA = doc.queryLayout(layoutAId);
-            var layoutB = doc.queryLayout(layoutBId);
-            if (!isNull(layoutA) && !isNull(layoutB)) {
-                return layoutA.getTabOrder() - layoutB.getTabOrder();
-            }
-        }
-
-        // sort by layout name or block name (ignoring * at start):
-        var blockTitleA = RBlockListQt.getBlockTitle(blockA);
-        var blockTitleB = RBlockListQt.getBlockTitle(blockB);
-        if (blockTitleA.startsWith("*")) {
-            blockTitleA = blockTitleA.substring(1);
-        }
-        if (blockTitleB.startsWith("*")) {
-            blockTitleB = blockTitleB.substring(1);
-        }
-
-        return Array.alphaNumericalSorter(blockTitleA, blockTitleB);
-    });
+    // add pen icon to item:
+    if (!isNull(this.currentItem)) {
+        this.currentItem.setIcon(BlockList.colEdit, BlockList.iconEdit[1]);
+    }
 };
 
 /**
  * Called when blocks are addded, edited or deleted. Updates the block list.
  */
 RBlockListQt.prototype.updateBlocks = function(documentInterface) {
+    this.currentItem = undefined;
+
+    RDebug.startTimer();
     this.di = documentInterface;
 
     var pos = this.verticalScrollBar().sliderPosition;
@@ -196,15 +222,15 @@ RBlockListQt.prototype.updateBlocks = function(documentInterface) {
 
     var doc = documentInterface.getDocument();
 
-    var blockNameCandidates = doc.getBlockNames();
+    var blockCandidateIds = doc.queryAllBlocks();
 
     var i;
 
     // filter block names:
-    var blockNames = [];
-    for (i=0; i<blockNameCandidates.length; i++) {
-        var blockNameCandidate = blockNameCandidates[i];
-        var blockCandidate = doc.queryBlockDirect(blockNameCandidate);
+    var blockIds = [];
+    for (i=0; i<blockCandidateIds.length; i++) {
+        var blockCandidateId = blockCandidateIds[i];
+        var blockCandidate = doc.queryBlockDirect(blockCandidateId);
         if (blockCandidate.isNull()) {
             continue;
         }
@@ -213,29 +239,30 @@ RBlockListQt.prototype.updateBlocks = function(documentInterface) {
             continue;
         }
 
-        blockNames.push(blockNameCandidate);
+        blockIds.push(blockCandidateId);
     }
 
-    this.sortBlockNames(blockNames);
+    blockIds = this.sortBlocks(doc, blockIds);
 
     var currentBlockId = doc.getCurrentBlockId();
-    var currentItem = undefined;
     var selectedItem = undefined;
-    for (i=0; i<blockNames.length; ++i) {
-        var blockName = blockNames[i];
-        var block = doc.queryBlockDirect(blockName);
+    for (i=0; i<blockIds.length; ++i) {
+        var blockId = blockIds[i];
+        var block = doc.queryBlockDirect(blockId);
         if (block.isNull()) {
             continue;
         }
 
-        var item = this.getBlockItem(block);
+        var blockName = block.getName();
+
+        var item = this.createBlockItem(block);
         this.addTopLevelItem(item);
         if (blockName===selectedBlockName) {
             selectedItem = item;
         }
 
         if (currentBlockId===block.getId()) {
-            currentItem = item;
+            this.currentItem = item;
         }
     }
 
@@ -248,6 +275,11 @@ RBlockListQt.prototype.updateBlocks = function(documentInterface) {
     }
 
     this.blockActivated();
+    RDebug.stopTimer("block list update");
+};
+
+RBlockListQt.prototype.sortBlocks = function(doc, blockIds) {
+    return doc.sortBlocks(blockIds);
 };
 
 RBlockListQt.getBlockTitle = function(block) {
@@ -259,7 +291,7 @@ RBlockListQt.getBlockTitle = function(block) {
     return title;
 };
 
-RBlockListQt.prototype.getBlockItem = function(block) {
+RBlockListQt.prototype.createBlockItem = function(block) {
     var item = new QTreeWidgetItem();
     var name = block.getName();
 
@@ -278,13 +310,13 @@ RBlockListQt.prototype.getBlockItem = function(block) {
 };
 
 RBlockListQt.prototype.updateItemIcons = function(item, block) {
-    var iconName = autoIconPath(BlockList.includeBasePath + "/Visible%1.svg".arg(Number(!block.isFrozen())));
-    item.setIcon(BlockList.colVisible, new QIcon(iconName));
+    //var iconName = autoIconPath(BlockList.includeBasePath + "/Visible%1.svg".arg(Number(!block.isFrozen())));
+    item.setIcon(BlockList.colVisible, BlockList.iconVisible[Number(!block.isFrozen())]);
 
     var doc = this.di.getDocument();
     var currentBlockId = doc.getCurrentBlockId();
-    iconName = autoIconPath(BlockList.includeBasePath + "/Edit%1.svg".arg(Number(block.getId()===currentBlockId)));
-    item.setIcon(BlockList.colEdit, new QIcon(iconName));
+    //iconName = autoIconPath(BlockList.includeBasePath + "/Edit%1.svg".arg(Number(block.getId()===currentBlockId)));
+    item.setIcon(BlockList.colEdit, BlockList.iconEdit[Number(block.getId()===currentBlockId)]);
 };
 
 /**
@@ -313,11 +345,11 @@ RBlockListQt.prototype.itemColumnClickedSlot = function(item, column) {
         block.setFrozen(!block.isFrozen());
         var op = new RModifyObjectOperation(block, false);
         this.di.applyOperation(op);
-        this.updateBlocks(this.di);
+        //this.updateBlocks(this.di);
     } else if (column===BlockList.colEdit) {
         this.setCurrentItem(item);
         this.editBlock();
-        this.updateBlocks(this.di);
+        //this.updateBlocks(this.di);
     }
 
     this.blockActivated();
@@ -360,7 +392,9 @@ RBlockListQt.prototype.blockActivated = function() {
 
     // prevent removal or insertion of current block:
     var insertable = currentBlock.getName().toLowerCase()!==blockName.toLowerCase() && !blockName.startsWith("*");
-    this.enableActions(insertable);
+    var renamable = document.getBlockId(blockName)!==document.getModelSpaceBlockId();
+    var removable = document.getBlockId(blockName)!==document.getModelSpaceBlockId();
+    this.enableActions(insertable, renamable, removable);
 };
 
 /**
@@ -370,12 +404,32 @@ RBlockListQt.prototype.getBlockName = function(item) {
     return item.data(BlockList.colName, Qt.UserRole);
 };
 
+/**
+ * \return The list item for the block with the given block name.
+ */
+RBlockListQt.getItem = function(widget, blockName) {
+    for (var i=0; i<widget.topLevelItemCount; i++) {
+        var item = widget.topLevelItem(i);
+        if (item.data(BlockList.colName, Qt.UserRole)===blockName) {
+            return item;
+        }
+    }
+    return undefined;
+};
+
 /**
  * Enable / disable buttons that edit, remove or insert the active block.
  */
-RBlockListQt.prototype.enableActions = function(enable) {
+RBlockListQt.prototype.enableActions = function(insertable, renamable, removable) {
+    if (isNull(renamable)) {
+        renamable = insertable;
+    }
+    if (isNull(removable)) {
+        removable = insertable;
+    }
+
     var override;
-    if (enable===true) {
+    if (insertable===true) {
         override = 1;
     }
     else {
@@ -384,13 +438,15 @@ RBlockListQt.prototype.enableActions = function(enable) {
 
     var action;
     action = RGuiAction.getByScriptFile("scripts/Block/ToggleBlockVisibility/ToggleBlockVisibility.js");
-    action.setEnabledOverride(enable, override);
+    action.setEnabledOverride(insertable, override);
     action = RGuiAction.getByScriptFile("scripts/Block/InsertBlock/InsertBlock.js");
-    action.setEnabledOverride(enable, override);
+    action.setEnabledOverride(insertable, override);
+
     action = RGuiAction.getByScriptFile("scripts/Block/RemoveBlock/RemoveBlock.js");
-    action.setEnabledOverride(enable, override);
+    action.setEnabledOverride(removable, removable ? 1 : 0);
+
     action = RGuiAction.getByScriptFile("scripts/Block/RenameBlock/RenameBlock.js");
-    action.setEnabledOverride(enable, override);
+    action.setEnabledOverride(renamable, renamable ? 1 : 0);
 };
 
 /**
@@ -424,6 +480,16 @@ BlockList.colVisible=0;
 BlockList.colEdit=1;
 BlockList.colName=2;
 
+BlockList.iconVisible = [
+    new QIcon(autoIconPath(BlockList.includeBasePath + "/Visible0.svg")),
+    new QIcon(autoIconPath(BlockList.includeBasePath + "/Visible1.svg"))
+];
+
+BlockList.iconEdit = [
+    new QIcon(autoIconPath(BlockList.includeBasePath + "/Edit0.svg")),
+    new QIcon(autoIconPath(BlockList.includeBasePath + "/Edit1.svg"))
+];
+
 BlockList.getPreferencesCategory = function() {
     return [ qsTr("Widgets"), qsTr("Block List") ];
 };
@@ -436,6 +502,18 @@ BlockList.applyPreferences = function(doc, mdiChild) {
     WidgetFactory.initList(blockList, "BlockList");
 };
 
+/**
+ * \return Active item.
+ */
+BlockList.getActiveItem = function() {
+    var appWin = RMainWindowQt.getMainWindow();
+    var blockList = appWin.findChild("BlockList");
+    if (isNull(blockList)) {
+        return undefined;
+    }
+
+    return blockList.getActiveItem();
+};
 
 /**
  * Shows / hides the block list.
